/* generates Allen CSPs on stdin
 * call:
 *    gencsp [-i n] [-n n] [-l n] [-c n] [-d n] [-s n] [-u] [-r file] [-x file]
 * where 
 * -i  specifies the number of instances to be generated (default:1)
 * -n  specifies the number of nodes (default: 3, min: 3, max: MAXCSP)
 * -l  specifies the labelsize (default: 6.5, min: 1, max: 13) 
 * -c  specifies the constraint density (default: 0.0, min: 0.0, max: 1.0)
 * -d  specifies the average degree of a node (default: 0, min: 0, max: MAXCSP)
 * -s  specifies the seed for the random function
 * -u  use uniform distribution (and exclude inconsistent networks)
 * -r  use only relations from file (uniform distribution)
 * -x  exclude the relations specified in the file (except definite ones)
 * -a  adds relations to labels to make the CSP consistent (van Beek)
 *
 */

const char *usagestring = "\
    gencsp [-i n] [-n n] [-l n] [-c n] [-d n] [-s n] [-u] [-r file] [-x file] 
       -i  specifies the number of instances to be generated (default: 1)
       -n  specifies the number of nodes (default: 3, min: 3, max: MAXCSP)
       -l  specifies the labelsize (default: 6.5, min: 1.0, max: 13.0) 
       -c  specifies the constraint density (default: 0.0, min: 0.0, max: 1.0)
       -d  specifies the avg degree of a node (default: 0.0, min: 0.0, max: MAXCSP)
       -s  specifies the seed for the random function
       -u  means to use a uniform distribution (excluding CSP with zero labs)
       -x file   specfies the relations to be excluded (except definite ones)
       -r file   restricts all relations to be the ones mentioned in the file
       -a  adds relations to labels to make the CSP consistent (van Beek)

       The -c and -d options override each other and are overriden by -u.
       If -r is specified, -l is ignored.";



#include "allen.h"
#include "allenio.h"
#include "allenop.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/times.h>
#include <sys/types.h>

/* switches */
int   swinstances = 1;
int   swnodes  = 3;
float swlabsiz = 6.5;
float swconsdensity = 0.0;
float swdegree = 0.0;
int swexclude = 0;
int swaddcons = 0;
char xalg[MAXSET];
int swrestrict = 0;
char ralg[MAXSET];
RELTYPE ralglist[MAXSET];
int restrictsize;
int swuniform = 0;
unsigned int   swseed = 0;

/* CSP to be generated */
RELTYPE csp[MAXCSP][MAXCSP];

/* type string */
char csptype[100];

/* random variables */
unsigned int randomseed;

void initrand(unsigned int seed)
{
  if (seed == 0)
    randomseed = time(NULL);
  else
    randomseed = seed;

  srand48(randomseed);
}


RELTYPE genlab(float avglabsiz, int uniform, int restrict, int exclude)
{
  long int thres;
  RELTYPE rel, shift;
  int onerelx;
  int i;
  int disjcnt;
  
  if (restrict)
    return(ralglist[abs(lrand48())%restrictsize]);

  if (uniform) {
    thres = 4999;
    onerelx = -1;
  } else {
    thres = ((int)((avglabsiz-1.0)*10000.0/12.0))-1;
    onerelx = abs(lrand48())%13;
  }

  do {
    shift = 1;
    i = 0;
    rel = 0;
    disjcnt = 1;

    do {
      if (i == onerelx) {
	rel |= shift;
	disjcnt++;
      }
      else
	if (abs(lrand48())%10000 <= thres) {
	  rel |= shift;
	  disjcnt++;
	}
      shift = shift << 1;
    } while (++i < 13);
  } while (swexclude && xalg[rel] && (disjcnt > 1));
  
  return(rel);
}

void gennet(RELTYPE csp[MAXCSP][MAXCSP],
	    char *typestring, 
	    int instcnt,
	    int cspsize, 
	    unsigned int seed,
	    float avglabsiz,
	    float avgcondense,
	    float avgdegree)
{
  register int i,j;
  long int thres;
  int incons;
  
  sprintf(typestring,"#%1d-N%1d-%c%1.2f-%c%1.2f-%s%sS%1u",
	  instcnt,cspsize,
	  (swrestrict ? 'R' : (swuniform ? 'U' : 'L')),
	  ((swrestrict || swuniform) ? 0.0 : avglabsiz),
	  (avgcondense != 0.0 ? 'C' : 'D'),
	  (avgcondense != 0.0 ? avgcondense : avgdegree),
	  (swexclude ? "X-" : ""),
	  (swaddcons ? "A-" : ""),
	  seed);
  
  if (avgcondense == 0.0) 
    avgcondense = (avgdegree*((float)cspsize))/((float)(cspsize*(cspsize-1)));

  thres = ((int)(avgcondense*10000.0))-1;

  do {
    incons = 0;
    for (i = 0; i < cspsize; i++)
      for (j = i+1; j < cspsize; j++)
	if (abs(lrand48())%10000 <= thres) {
	  if (!(csp[i][j] = genlab(avglabsiz,swuniform,swrestrict,swexclude))) {
	    i = cspsize;
	    j = cspsize;
	    incons = 1;
	    break;
	  }
	} else
	  csp[i][j] = DALL;
  } while (incons);
}

void makeconsistent(RELTYPE csp[MAXCSP][MAXCSP], int nodes)
{
  register int i,j;
  int e[MAXCSP], s[MAXCSP];

  /* code from van Beek */
  for (i = 0; i < nodes; i++) {
    s[i] = ((abs(lrand48()))%71)+1;
    e[i] = s[i] + ((abs(lrand48()))%17)+1;
  }

  for (i = 0; i < nodes; i++)
    for (j = i+1; j < nodes; j++)
      if      (i == j)			        csp [i][j] |= REQ;
      else if (e[i] <  s[j])			csp [i][j] |= RLT ;
      else if (s[i] >  e[j])			csp [i][j] |= RGT;
      else if (e[i] == s[j])			csp [i][j] |= RM ;
      else if (s[i] == e[j])			csp [i][j] |= RMI;
      else if (s[i] >  s[j] && e[i] <  e[j])	csp [i][j] |= RD ;
      else if (s[i] <  s[j] && e[i] >  e[j])	csp [i][j] |= RDI;
      else if (s[i] == s[j] && e[i] <  e[j])	csp [i][j] |= RS ;
      else if (s[i] == s[j] && e[i] >  e[j])	csp [i][j] |= RSI;
      else if (s[i] >  s[j] && e[i] == e[j])	csp [i][j] |= RF ;
      else if (s[i] <  s[j] && e[i] == e[j])	csp [i][j] |= RFI;
      else if (s[i] == s[j] && e[i] == e[j])	csp [i][j] |= REQ;
      else if (s[i] <  s[j] && e[i] <  e[j]
	       && e[i] >  s[j])	csp [i][j] |= RO ;
      else if (s[i] >  s[j] && e[i] >  e[j]
	       && s[i] <  e[j])	csp [i][j] |= ROI;
      else {
	fprintf (stderr,"error:s[%d] = %d, e[%d] = %d\n",i,s[i],i,e[i]);
	fprintf (stderr,"      s[%d] = %d, e[%d] = %d\n",j,s[j],j,e[j]);
      }
}

    
void usage()
{
  fprintf(stderr,"%s\n",usagestring);
  exit(-1);
}

unsigned int readintarg(int *xarg, char **argv)
{
  unsigned int result;

  if (argv[*xarg][2] == '\0') {
    (*xarg)++;
    if (sscanf(argv[*xarg],"%d",&result) != 1) usage();
  } else {
    if (sscanf(&argv[*xarg][2],"%d",&result) != 1) usage();
  }
  return(result);
}

float readfloatarg(int *xarg, char **argv)
{
  float result;

  if (argv[*xarg][2] == '\0') {
    (*xarg)++;
    if (sscanf(argv[*xarg],"%f",&result) != 1) usage();
  } else {
    if (sscanf(&argv[*xarg][2],"%f",&result) != 1) usage();
  }
  return(result);
}

void main(int argc, char **argv)
{
  int xarg;
  register int i, j;
  FILE *algfile;

  xarg = 1;

  while (xarg < argc) {
    if (argv[xarg][0] != '-') usage();
    switch(argv[xarg][1]) {
    case 'i':
      swinstances = readintarg(&xarg,argv);
      break;
    case 'n':
      swnodes = readintarg(&xarg,argv);
      break;
    case 'l':
      swlabsiz = readfloatarg(&xarg,argv);
      break;
    case 'c':
      swconsdensity = readfloatarg(&xarg,argv);
      if (swconsdensity != 0.0) swdegree = 0.0;
      break;
    case 'd':
      swdegree = readfloatarg(&xarg,argv);
      if (swdegree != 0.0) swconsdensity = 0.0;
      break;
    case 's':
      swseed = readintarg(&xarg,argv);
      break;
    case 'u':
      swuniform = 1;
      break;
    case 'a':
      swaddcons = 1;
      break;
    case 'r':
      swrestrict = 1;
      if ((algfile = fopen(argv[++xarg],"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",argv[xarg]);
	exit(-1);
      }
      if (!(parsealg(algfile,ralg))) {
	fprintf(stderr,"\n*** Error in reading rel set from '%s'\n",argv[xarg]);
	exit(-1);
      }
      fclose(algfile);
      break;
    case 'x':
      swexclude = 1;
      if ((algfile = fopen(argv[++xarg],"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",argv[xarg]);
	exit(-1);
      }
      if (!(parsealg(algfile,ralg))) {
	fprintf(stderr,"\n*** Error in reading rel set from '%s'\n",argv[xarg]);
	exit(-1);
      }
      fclose(algfile);
      break;
    default:
      usage();
    }
    xarg++;
  }

  if ((swinstances < 1) || (swnodes < 3) || 
      (swnodes > MAXCSP) ||
      ((swlabsiz < 1.0) && !swrestrict) || 
      ((swconsdensity < 0.1) && (swdegree < 0.1) && !swuniform))
    usage();

  initrand(swseed);
  
  restrictsize = 0;
  if (swrestrict) {
    for (i=1; i<MAXSET; i++) 
      if (ralg[i] && 
	  ((i != DALL) || (swconsdensity == 1.0)) && 
	   (!swexclude || !xalg[i]))
	ralglist[restrictsize++] = i;
    if (restrictsize == 0) {
      fprintf(stderr,"\n*** Restrict-set has size 0!\n");
      exit(-1);
    }
  }

  for (i = 1; i <= swinstances; i++) {
    gennet(csp, csptype, i, swnodes, randomseed, 
	   swlabsiz, swconsdensity, swdegree);
    if (swaddcons) makeconsistent(csp, swnodes);
    printcsp(stdout, swnodes-1, csptype, csp);
  }
  exit(0);
}
