/* solves an Allen CSP 
 * call it
 *    solve [-dvrqs] [file1 file2 ...]
 * where 
 * -d means debugging info,
 * -v means verbose, 
 * -q means quite (only give a return value of 0 if consistent, 1 if incons.,
 *    (in this case only one CSP is processed!)
 * -s means a summary of all CSP solving attempts should be printed
 * -n means to restrict summary to negative (inconsistent) cases
 * -p means to restrict summary to positive (consistent) cases
 * -b means brief one-line summary
 * -a another format of a brief one-line summary
 * -c <file> means to use the coverage info in file to split relations (default
 *         is to split into atomic relations)
 * -r means to use random variable ordering  (ordered by 
 *         heuristic/cardinality is default)
 * -w use weighted queue scheme in pc from van Beek (simple marking is default)
 * -e use the environment constrainedness to select next cand (van Beek)
 * -f use fixed ordering of variables (dynamic is default)
 * -t <n> timeout after n ticks (return 0)
 * -o <file> solve only those CSP that a tagged with zero (should be a log-file)
 * filei should contain one (or more) Allen CSP(s).
 * The first line should contain the maximum nodeid (<MAXCSP) and a 
 * a type string.
 * The following lines should contain
 *   nodeid1 nodeid2 relation
 * terminated by a line with a dot in the first column.
 * If more than file is specified, all are processed sequentially.
 * If no file is specified, we read from stdin.
 */

#define NV_UPPER_LIMIT 2000000 /* if more than this # of nodes visited, fail */

const char *usagestring = "\
    solve [-dvrqsbpnwa] [-c file] [file1 file2 ...]      solves Allen CSPs
    -c <file> means to use the coverage info in file to split relations
    -d    give debugging info on stderr
    -v    give verbose output on stdout after solving a CSP
    -q    return 0 if (last) CSP was consistent (suppresses -v) 
    -s    return summary info on stdout (even if -q)
    -n    means to restrict summary to negative (inconsistent) cases
    -p    means to restrict summary to positive (consistent) cases
    -b    return one-line summary info on stdout (even if -q)
    -a    one-line summary (other format)
    -r    means to use random variable ordering  
    -w    use weighted queue scheme from van Beek
    -e    use environmemt constrainedness to select next node (van Beek)
    -f    fixed ordering of variables 
    -t <n> timeout after n ticks (return 0)
    -o <file> solve only those CSP that a tagged with zero 
              (should be a log-file)

    filei are input files, if none specified use stdin. More than one CSP 
          can be specified in each file";


#include "allen.h"
#include "allenio.h"
#include "allenfop.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/times.h>
#include <sys/types.h>



/* maximal number of trials */
#define MAXTRIAL 20000

/* switches */
int swdebug;
int swverbose;
int swquite;
int swsummary;
int swbrief;
int swabrief;
int swpositive;
int swnegative;
int swposneg;
int swsplitinfo;
int swrandomorder;
int swwqueue;
int swenvcons;
int swfixed;
int swtimeout;
int swsolveonly;

/* statistical info */
int maxnodeid;
char csptype[100];
int consistent;
int cntconsistent;
int pathconsistent;
int cntpathconsistent;
int pcops;
int pcits;
int nodesvisited;
int cputicks;
int trial;
int rectrial;
long int tpcops[MAXTRIAL];
long int tpcits[MAXTRIAL];
long int tnodesvisited[MAXTRIAL];
long int tcputicks[MAXTRIAL];
int cumcputicks;
int cntsize;
int size;
RELTYPE splitinfo[MAXSET][14];

/* CSP */
RELTYPE csp[MAXCSP][MAXCSP];

/* constrainedness of variable */
int consval[MAXCSP][MAXCSP];

/* files */
FILE *infile;
char infilename[200];
FILE *splitfile;
char splitfilename[200];
FILE *solveonlyfile;
char solveonlyname[200];

/* timing info */
struct tms local_before, local_after, global_before, global_after;


/* random variable */
unsigned int randomseed = 0;

void initrand(unsigned int seed)
{
  if (seed == 0)
    randomseed = time(NULL);
  else
    randomseed = seed;

  srand48(randomseed);
}

int path_cons(RELTYPE csp[MAXCSP][MAXCSP], int maxnodeid, int node1, int node2)
{
  if (swwqueue) 
    return(path_cons_vanbeek(csp,maxnodeid,node1,node2));
  else
    return(path_cons1(csp,maxnodeid,node1,node2));
}

/* compute the constrainedness of the environment of an edge */
int envcons(RELTYPE csp[MAXCSP][MAXCSP], int maxnodeid, register int i, register int j)
{
  register int k;
  register int w = 0;
  

  for (k = 0; k <= maxnodeid; k++) {
    if ((k != i) && (k != j)) {
      if (k < i) w += FVAL(csp[k][i]);
      else w += FVAL(FINV(csp[i][k]));
      if (k < j) w += FVAL(csp[k][j]);
      else w += FVAL(FINV(csp[j][k]));
    }
  }
  return(w);
}

/* check whether csp is already solved,
 * if so return 1,
 * else return 0 and give a relation that can be split
 */
int solved_csp(RELTYPE csp[MAXCSP][MAXCSP], int maxnodeid, int *node1, int *node2) 
{
  register int i,j;
  register int temp, minval;
  register RELTYPE r;
  int defaultval = ((MAXCSP+1)*(MAXCSP+1)*20000);

  minval = defaultval;
  *node1 = -1;
  *node2 = -1;
  for (i=0; i <= maxnodeid; i++)
    for (j=i+1; j <= maxnodeid; j++) {
      if (csp[j][i]) {
	r = csp[i][j];
	if (swsplitinfo) {
	  if (splitinfo[r][0] == 1) temp=0;
	  else temp = 
	    FVAL(r)*14+(splitinfo[r][0]); 
	} else {
	  temp =  (FVAL(r));
	  if ((r == REQ) || (r == RGT) || (r == RLT) || (r == RD) || 
	      (r == RDI) || (r == RO) || (r == ROI) || (r == ROI) ||
	      (r == RM) || (r == RMI) || (r == RS) || (r == RSI) ||
	      (r == RF) || (r == RFI) || (r ==DALL))
	    temp = 0;
	}
	if (temp == 0)
	  csp[j][i] = 0;
	else
	  if (temp > 0) {
	    if (swfixed) temp = consval[i][j];
	    else if (swenvcons) temp = 400*envcons(csp, maxnodeid, i,j)+temp;
	    if ((temp < minval) && (temp > 0)) {
	      minval = temp;
	      *node1 = i;
	      *node2 = j;
	    }
	  }
      }
    }
  if (minval == defaultval) return(1);
  else 
    if (!swrandomorder || swfixed) return(0);
    else {
      if (randomseed == 0) 
	initrand(0);
      temp = abs(lrand48())%(maxnodeid*2);
      do {
	for (i=0; i <= maxnodeid; i++)
	  for (j=i+1; j <= maxnodeid; j++) 
	    if (csp[j][i]) 
	      if (--temp <= 0) {
		*node1 = i;
		*node2 = j;
		return(0);
	      }
      } while (1);
    }
}

/* compute initial cons val */
init_consval(RELTYPE csp[MAXCSP][MAXCSP], int maxnodeid)
{
  register RELTYPE r;
  register int i,j, temp;

  for (i=0; i <= maxnodeid; i++)
    for (j=i+1; j <= maxnodeid; j++) {
	r = csp[i][j];
	if (swsplitinfo) {
	  if (splitinfo[r][0] == 1) temp=0;
	  else temp = 
	    (FVAL(r)*14+(splitinfo[r][0])); 
	} else {
	  temp = (FVAL(r));
	  if ((r == REQ) || (r == RGT) || (r == RLT) || (r == RD) || 
	      (r == RDI) || (r == RO) || (r == ROI) || (r == ROI) ||
	      (r == RM) || (r == RMI) || (r == RS) || (r == RSI) ||
	      (r == RF) || (r == RFI) || (r ==DALL))
	    temp = 0;
	}
	if (temp <= 0) consval[i][j] = 0;
	else {
	  if (swenvcons) temp = 400*envcons(csp, maxnodeid, i,j)+temp;
	  if (swrandomorder) temp = abs(abs(lrand48())%(temp*100));
	  consval[i][j] = temp;
	}
      }
}


/* generate a "split" */
void gen_split(RELTYPE r, RELTYPE split[14])
{
  int i;
  
  if (swsplitinfo) {
    for (i=1;i<=13;i++) split[i-1] = splitinfo[r][i];
    split[5] = 0;
  } else {
    i = 0;
    if (r & RO) split[i++] = RO;
    if (r & ROI) split[i++] = ROI;
    if (r & RD) split[i++] = RD;
    if (r & RLT) split[i++] = RLT;
    if (r & RGT) split[i++] = RGT;
    if (r & RDI) split[i++] = RDI;
    if (r & RM) split[i++] = RM;
    if (r & RMI) split[i++] = RMI;
    if (r & RS) split[i++] = RS;
    if (r & RSI) split[i++] = RSI;
    if (r & RF) split[i++] = RF;
    if (r & RFI) split[i++] = RFI;
    if (r & REQ) split[i++] = REQ;
    split[i] = 0;
  }
}

/* checks CSP for global consistency */
int global_cons(RELTYPE csp[MAXCSP][MAXCSP], int maxnodeid, int node1, int node2)
{
  RELTYPE newcsp[MAXCSP][MAXCSP];
  RELTYPE split[14];
  int splitnode1, splitnode2, s;
  register int i,j;
  int node;
  struct tms local_time;

  if (node1 < 0) {
    for (i = 0; i <= maxnodeid; i++)
      for (j = i+1; j <= maxnodeid; j++) {
	csp[j][i] = 1;
	if (csp[i][j] == 0) return(0);
      }
    /* mark all nodes that we don't have to split */
    solved_csp(csp,maxnodeid,&splitnode1,&splitnode2);
  }

  if (swtimeout) {
    times(&local_time);
    if (swtimeout < (local_time.tms_utime - local_before.tms_utime)) {
      if (swdebug) {
	putc('T',stderr);
	fflush(stderr);
      }
      return(0);
    }
  }

  nodesvisited++;
  node = nodesvisited;

  if (node > NV_UPPER_LIMIT) return(0);

  if (swdebug) {
    fprintf(stderr,"D%1d",node);
    fflush(stderr);
  }

  for (i = 0; i <= maxnodeid; i++)
    for (j = 0; j <= maxnodeid; j++)
      newcsp[i][j] = csp[i][j];
  if (path_cons(newcsp,maxnodeid,node1,node2)) {
    if (node1 < 0) {
      pathconsistent = 1;
      if (swfixed) init_consval(csp,maxnodeid);
    }
    if (solved_csp(newcsp,maxnodeid,&splitnode1,&splitnode2)) {
      return(1);
    } else {
      gen_split(newcsp[splitnode1][splitnode2],split);
      for (s = 0; s < 14; s++) {
	if (split[s] != 0) {
	  newcsp[splitnode1][splitnode2] = split[s];
	  if (global_cons(newcsp,maxnodeid,splitnode1,splitnode2)) {
	    if (swdebug) {
	      fprintf(stderr,"U%1d",node);
	      fflush(stderr);
	    }
	    return(1);
	  }
	} else break;
      }
    }
  }
  if (swdebug) {
    fprintf(stderr,"u%1d",node);
    fflush(stderr);
  }
  return(0);
}


void usage()
{
  fprintf(stderr,"%s\n",usagestring);
  exit(-1);
}

int int_compare(long int *i1, long int *i2)
{
  return(((*i1 < *i2) ? -1 
	  : ((*i1 == *i2) ? 0 : 1)));
}

void print_summary(int brief, int abrief)
{
  float avg, med, seventy, ninety, ninetynine;
  register int i;

  if (!brief && !abrief) printf("SUMMARY INFORMATION:\n");
  if (trial < 1) {
    printf("  No CSP given.\n");
    return;
  }

  if (brief || abrief) printf("%.24s ",csptype);
  else 
    printf("  CSP size=%5.1f (%s), last type=%s\n",
	   ((float)cntsize)/((float)(trial)),
	   ((size == 0) ? "avg." : "all"),
	   csptype);

  if (brief || abrief) printf("%5.1f:%5.1f ",
		    (((float)cntconsistent)/((float)trial))*100.0,
		    (((float)cntpathconsistent)/((float)trial))*100.0);
  else 
    printf("  Consistent: path=%6.2f%%, global=%6.2f%%\n",
	   (((float)cntpathconsistent)/((float)trial))*100.0,
	   (((float)cntconsistent)/((float)trial))*100.0);

  if (swpositive) {
    if (brief || abrief) printf("C ");
    else printf("  Statistics on CONSISTENT cases only:\n");
  }
  else if (swnegative) {
    if (brief || abrief) printf("I ");
    else printf("  Statistics on INCONSISTENT cases only:\n");
  }

  trial = rectrial;

  if (trial <= 0) {
    if (brief || abrief) printf(" -------\n");
    else printf("  *** no cases\n");
    return;
  }
  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tcputicks[i]);
  avg = avg/((float)trial);
  qsort(tcputicks,trial,sizeof(long int),&int_compare);
  if (trial == 1) {
    med = (float)tcputicks[0];
    seventy = (float)tcputicks[0];
    ninety = (float)tcputicks[0];
    ninetynine = (float)tcputicks[0];
  } else {
    med = ((float)(tcputicks[(trial/2)-1]+tcputicks[trial/2]))/2.0;
    seventy = ((float)(tcputicks[((trial*7)/10)-1]+
		       tcputicks[((trial*7)/10)]))/2.0;
    ninety = ((float)(tcputicks[((trial*9)/10)-1]+
		       tcputicks[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tcputicks[((trial*99)/100)-1]+
		       tcputicks[((trial*99)/100)]))/2.0;
  }
  if (brief) printf("%4.1f %4.1f %4.1f ",avg,med,ninety);
  else if (abrief)
    printf("%4.1f %4.1f %4.1f %4.1f %4.1f ",avg,med,seventy,ninety,ninetynine);
  else
    printf("  CPU time:   avg.=%6.1ft percentile= %6.1ft %6.1ft %6.1ft %6.1ft\n",
	   avg, med, seventy, ninety, ninetynine);

  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tpcops[i]);
  avg = avg/((float)trial);
  qsort(tpcops,trial,sizeof(long int),&int_compare);
  if (trial == 1) {
    med = (float)tpcops[0];
    seventy = (float)tpcops[0];
    ninety = (float)tpcops[0];
    ninetynine = (float)tpcops[0];
  } else {
    med = ((float)(tpcops[(trial/2)-1]+tpcops[trial/2]))/2.0;
    seventy = ((float)(tpcops[((trial*7)/10)-1]+
		       tpcops[((trial*7)/10)]))/2.0;
    ninety = ((float)(tpcops[((trial*9)/10)-1]+
		       tpcops[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tpcops[((trial*99)/100)-1]+
		       tpcops[((trial*99)/100)]))/2.0;
  }
  if (!brief && !abrief)
    printf("  PC ops:     avg.=%7.0f percentile= %7.0f %7.0f %7.0f %7.0f\n",
	   avg, med, seventy, ninety, ninetynine);

  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tpcits[i]);
  avg = avg/((float)trial);
  qsort(tpcits,trial,sizeof(long int),&int_compare);
  if (trial == 1) {
    med = (float)tpcits[0];
    seventy = (float)tpcits[0];
    ninety = (float)tpcits[0];
    ninetynine = (float)tpcits[0];
  } else {
    med = ((float)(tpcits[(trial/2)-1]+tpcits[trial/2]))/2.0;
    seventy = ((float)(tpcits[((trial*7)/10)-1]+
		       tpcits[((trial*7)/10)]))/2.0;
    ninety = ((float)(tpcits[((trial*9)/10)-1]+
		       tpcits[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tpcits[((trial*99)/100)-1]+
		       tpcits[((trial*99)/100)]))/2.0;
  }
  if (brief)
    printf("/ %4.1f %4.1f %4.1f ",avg, med,  ninety);
  else
    if (!abrief)
      printf("  PC its:     avg.=%7.1f percentile= %7.1f %7.1f %7.1f %7.1f\n",
	     avg, med, seventy, ninety, ninetynine);

  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tnodesvisited[i]);
  avg = avg/((float)trial);
  qsort(tnodesvisited,trial,sizeof(long int),&int_compare);
  
  if (trial == 1) {
    med = (float)tnodesvisited[0];
    seventy = (float)tnodesvisited[0];
    ninety = (float)tnodesvisited[0];
    ninetynine = (float)tnodesvisited[0];
  } else {
    med = ((float)(tnodesvisited[(trial/2)-1]+tnodesvisited[trial/2]))/2.0;
    seventy = ((float)(tnodesvisited[((trial*7)/10)-1]+
		       tnodesvisited[((trial*7)/10)]))/2.0;
    ninety = ((float)(tnodesvisited[((trial*9)/10)-1]+
		       tnodesvisited[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tnodesvisited[((trial*99)/100)-1]+
		       tnodesvisited[((trial*99)/100)]))/2.0;
  }
  if (brief)
    printf("/ %4.1f %4.1f %4.1f",
	   avg, med, ninety);
  else if (abrief)
    printf("/ %4.1f %4.1f %4.1f %4.1f %4.1f",
	   avg, med, seventy, ninety, ninetynine);
  else
    printf("  Nodes v.:   avg.=%7.1f percentile= %7.1f %7.1f %7.1f %7.1f\n",
	   avg, med, seventy, ninety, ninetynine);
  if (brief || abrief)
    printf("\n");
}

void readsplitfile(char *s)
{
  int i,j,inp;

  if ((splitfile = fopen(s,"r")) == NULL) {
    fprintf(stderr, "Relation-cover file '%s' not found\n",s);
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<14; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "Format error in relation-cover file '%s'\n",s);
	exit(-1);
      } else {
	splitinfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "No eof in relation-cover file '%s'\n",s);
      	exit(-1);
      }
}

void getnextcsp(char csp[100])
{
  int d;
  char line[256];

  d = 1;
  while ((d != 0) && (!feof(solveonlyfile))) {
    fgets(line,255,solveonlyfile);
    if (sscanf(line,"%s %d",csp,&d) != 2) d = 1;
    if ((d == 0) && (csp[strlen(csp)-1] == ':')) 
      csp[strlen(csp)-1] = '\0';
    if (swdebug) fprintf(stderr,"SO: %s -> %s %d\n",line,csp,d);
  }
  if (feof(solveonlyfile)) csp[0] = '\0';
}

unsigned int readintarg(int *xsw, int *xarg, char **argv)
{
  unsigned int result;

  if (argv[*xarg][++(*xsw)] == '\0') {
    (*xarg)++;
    if (sscanf(argv[*xarg],"%d",&result) != 1) usage();
  } else {
    if (sscanf(&argv[*xarg][*xsw],"%d",&result) != 1) usage();
  }
  *xsw = strlen(argv[*xarg])-1;
  return(result);
}

void main(int argc, char **argv)
{
  int xarg, xsw;
  int result;
  char nextcsp[100];

  swdebug = 0;
  swverbose = 0;
  swquite = 0;
  swsummary = 0;
  swpositive = 0;
  swnegative = 0;
  swposneg = 0;
  swsplitinfo = 0;
  swrandomorder = 0;
  swwqueue = 0;
  swenvcons = 0;
  swfixed = 0;
  swtimeout = 0;
  swsolveonly = 0;

  xarg = 1;
  while ((xarg < argc) && (argv[xarg][0] == '-')) {
    xsw = 1;
    while (argv[xarg][xsw] != '\0') {
      switch (argv[xarg][xsw]) {
      case 'c':
	swsplitinfo = 1;
	if (argv[xarg][++xsw] == '\0') {
	  strcpy(splitfilename, argv[++xarg]);
	} else {
	  strcpy(splitfilename, &argv[xarg][xsw]);
	}
	xsw = strlen(argv[xarg])-1;
	readsplitfile(splitfilename);
	break;
      case 'o':
	swsolveonly = 1;
	if (argv[xarg][++xsw] == '\0') {
	  strcpy(solveonlyname, argv[++xarg]);
	} else {
	  strcpy(solveonlyname, &argv[xarg][xsw]);
	}
	xsw = strlen(argv[xarg])-1;
	if ((solveonlyfile = fopen(solveonlyname,"r")) == NULL) {
	  fprintf(stderr,"Cannot open 'solve-only' file '%s'\n",solveonlyname);
	  exit(-1);
	}
	break;
      case 'd': 
	swdebug = 1;
	break;
      case 'v':
	swverbose = 1;
	break;
      case 'q': 
	swquite = 1;
	break;
      case 's':
	swsummary = 1;
	break;
      case 'p':
	swpositive = 1;
	swnegative = 0;
	swsummary = 1;
	swposneg = 1;
	break;
      case 'n':
	swpositive = 0;
	swnegative = 1;
	swsummary = 1;
	swposneg = 1;
	break;
      case 'b':
	swbrief = 1;
	swabrief = 0;
	swsummary = 1;
	break;
      case 'a':
	swbrief = 0;
	swabrief = 1;
	swsummary = 1;
	break;
      case 'r':
	swrandomorder = 1;
	break;
      case 'w':
	swwqueue = 1;
	break;
      case 'e':
	swenvcons = 1;
	break;
      case 'f':
	swfixed = 1;
	break;
      case 't':
	swtimeout = readintarg(&xsw,&xarg,argv);
	break;
      default:
	usage();
      }
      xsw++;
    }
    xarg++;
  }

  if (xarg >= argc) infile = stdin;
  else infile = NULL;

  if (swdebug) {
    putc('I',stderr);
    fflush(stderr);
  }
  initfop();
  if (swdebug) {
    putc('!',stderr);
    fflush(stderr);
  }

  trial = 0;
  rectrial = 0;
  cntpathconsistent = 0;
  cntconsistent = 0;
  cntsize = 0;
  size = -1;

  times(&global_before);

  do {
    if (infile == NULL) {
      strcpy(infilename,argv[xarg]);
      if ((infile = fopen(infilename,"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",infilename);
	exit(-1);
      }
    }

    while (!feof(infile)) {

      if (swsolveonly) getnextcsp(nextcsp);
      if (swdebug) fprintf(stderr,"Next: %s\n",nextcsp);

      do {

	result = parsecsp(infile,MAXCSP-1,&maxnodeid,csptype,csp);
	if (swdebug) fprintf(stderr,"Check: %s\n",csptype);

	if (size == -1) size = maxnodeid + 1;
	if (size != maxnodeid + 1) size = 0;
	if (result == -1) exit(-1);

      } while ((swsolveonly) &&
	       (strcmp(nextcsp,csptype) != 0) &&
	       (result != 0));

      if (result == 0) break;
      
      if ((trial >= MAXTRIAL) && (swsummary)) {
	fprintf(stderr,"\n*** More than %d CSPs given! Quiting now\n",MAXTRIAL);
	times(&global_after);
	cumcputicks = global_after.tms_utime - global_before.tms_utime;
	print_summary(swbrief,swabrief);
	if (swquite) exit(!consistent);
	else exit(0);
      }

      pcops = 0;
      pcits = 0;
      nodesvisited = 0;
      pathconsistent = 0;
      if (swdebug) {
	fprintf(stderr,"\nSolving %s:\n", csptype);
	fflush(stderr);
      }

      /* check global consistency */
      times(&local_before);
      consistent = global_cons(csp,maxnodeid,-1,-1);
      
      if (swdebug) {
	putc('\n',stderr);
	if (swquite) 
	  fprintf(stderr,"... solved. path-cons=%d, cons=%d\n",
		  pathconsistent,consistent);
	fflush(stderr);
      }

      /* statistics */
      times(&local_after);
      cputicks = local_after.tms_utime - local_before.tms_utime;
      if (consistent) cntconsistent++;
      if (pathconsistent) cntpathconsistent++;
      cntsize += maxnodeid+1;
      if (!swposneg || 
	  (consistent && swpositive) || 
	  (!consistent && swnegative)) {
	tpcops[rectrial] = pcops;
	tpcits[rectrial] = pcits;
	tnodesvisited[rectrial] = nodesvisited;
	tcputicks[rectrial] = cputicks;
	rectrial++;
      }

      /* output result */
      if (!swquite) {
	if (!swverbose) 
	  printf("%s: %d\n", csptype, consistent);
	else
	  printf("%s: %d  CPU=%dt, PC-ops=%d, PC-it=%d, Nodes=%d\n",
		 csptype, consistent,
		 cputicks, pcops, pcits, nodesvisited);
      }

      trial++;
    }
    xarg++;
    if (infile != stdin) fclose(infile);
    infile = NULL;
  } while (xarg < argc);

  times(&global_after);
  cumcputicks = global_after.tms_utime - global_before.tms_utime;
  if (swsummary) print_summary(swbrief,swabrief);
  if (swquite) exit(!consistent);
  else exit(0);
}    
